unit ifgdiplus;

interface

uses ImgFilter;

var
  GdiplusEnabled: Boolean;

type
  TGdiplusFilter = class(TImgFilter)
    constructor Create(Alloc: TImgFilterMemAlloc); override;
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal;
        var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
        override;
  end;

  TGdiplusSaveImageFormat = (gsifPNG, gsifJPEG);

procedure GdiplusSaveImage(const FileName: WideString;
    Format: TGdiplusSaveImageFormat; const BitmapInfo: Pointer;
    BitmapData: PChar);

implementation

uses Windows, ActiveX, GdiplusSub, NTypes, NLib;

const
  PropertyTagGpsLatitudeRef = $0001;
  PropertyTagGpsLatitude = $0002;
  PropertyTagGpsLongitudeRef = $0003;
  PropertyTagGpsLongitude = $0004;
  PropertyTagImageDescription = $010e;
  PropertyTagEquipMake = $010f;
  PropertyTagEquipModel = $0110;
  PropertyTagOrientation = $0112;
  PropertyTagSoftwareUsed = $0131;
  PropertyTagDateTime = $0132;
  PropertyTagArtist = $013b;
  PropertyTagCopyright = $8298;
  PropertyTagICCProfile = $8773;

  ErrorMessages: array [0..21] of string = (
    'Ok',
    'Generic error',
    'Invalid parameter',
    'Out of memory',
    'Object busy',
    'Insufficient buffer',
    'Not implemented',
    'Win32 error',
    'Wrong state',
    'Aborted',
    'File not found',
    'Value overflow',
    'Access denied',
    'Unknown image format',
    'Font family not found',
    'Font style not found',
    'Not TrueType font',
    'Unsupported GDI+ version',
    'GDI+ not initialized',
    'Property not found',
    'Property not supported',
    'Profile not found'
  );

procedure GdipCheck(Status: TGpStatus);
var
  msg: string;
begin
  if Status = gpstOk then Exit;

  if Ord(Status) < Length(ErrorMessages) then
    msg := ErrorMessages[Ord(Status)]
  else
    msg := 'Unknown error (' + IStr(Ord(Status)) + ')';
  raise EImgFilterError.Create('GDI+ error: ' + msg);
end;

constructor TGdiplusFilter.Create(Alloc: TImgFilterMemAlloc);
begin
  inherited;
  FName := 'Internal GDI+ decoder';
end;

function TGdiplusFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
var
  s: string;
  len: Cardinal;
begin
  SetLength(s, 16);
  if BufSize > Cardinal(Length(s)) then
    len := Length(s)
  else
    len := BufSize;
  Move(Buf^, s[1], len);

  Result :=
    (Copy(s, 1, 2) = #$ff#$d8) or // JPEG
    (Copy(s, 1, 8) = #$89'PNG'#$0d#$0a#$1a#$0a) or // PNG
    (Copy(s, 1, 6) = 'GIF87a') or (Copy(s, 1, 6) = 'GIF89a') or // GIF
    (Copy(s, 1, 4) = 'MM'#$00#$2a) or (Copy(s, 1, 4) = 'II'#$2a#$00); // TIFF
end;

function GetPropertyItem(Image: Pointer; Id: PROPID; var Item: PPropertyItem):
    UINT;
begin
  if GdipGetPropertyItemSize(Image, Id, Result) = gpstOk then
  begin
    GetMem(Item, Result);
    GdipCheck(GdipGetPropertyItem(Image, Id, Result, Item^));
  end
  else begin
    Item := nil;
    Result := 0;
  end;
end;

function GetProperty(Image: Pointer; Id: PROPID): string;
type
  PCardinalArray = ^TCardinalArray;
  TCardinalArray = array[0..5] of Cardinal;

  function FormatGpsPos(ref: PChar; pos: PCardinalArray): string;
  var
    posd: Double;
  begin
    Result := Format('{0} {1}/{2} {3}/{4} {5}/{6}',
        [string(ref), pos[0], pos[1], pos[2], pos[3], pos[4], pos[5]]);
    if pos[1]*pos[3]*pos[5] = 0 then
      Exit;
    posd := pos[0]/pos[1] + pos[2]/pos[3]/60.0 + pos[4]/pos[5]/3600.0;
    if (ref = 'S') or (ref = 'W') then
      posd := -posd;
    Result := Result + ' (' + FStr(posd, 6) + ')';
  end;

  function GetICCProfileDesc(buf: PChar; size: Cardinal): string;
  var
    tagcount, ofs, len: Cardinal;
    p, pp: PChar;
    ws: WideString;
    i: Integer;
  begin
    Result := '';
    // header + tag count + one tag table
    if size <= (128 + 4 + 12) then Exit;

    p := buf;
    Inc(p, 128);
    tagcount := Be32ToH(PUInt32(p)^);
    Inc(p, 4);

    while (tagcount > 0) and (p < @buf[size - 12]) do
    begin
      if not ((p[0] = 'd') and (p[1] = 'e') and (p[2] = 's') and (p[3]='c')) then
      begin
        Inc(p, 12);
        Dec(tagcount);
        Continue;
      end;

      Inc(p, 4);
      ofs := Be32ToH(PUInt32(p)^);
      if ofs >= (size - 12) then Exit;

      p := @buf[ofs];

      if (p[0]='d') and (p[1]='e') and (p[2]='s') and (p[3]='c') then
      begin
        Inc(p, 8);
        len := Be32ToH(PUInt32(p)^);
        Inc(p, 4);
        if PChar(@p[len]) > @buf[size] then Exit;
        SetString(Result, p, len);
        Exit;
      end;

      if (p[0]='m') and (p[1]='l') and (p[2]='u') and (p[3]='c') then
      begin
        pp := p;
        Inc(pp, 20);
        if pp > @buf[size - 8] then Exit;
        len := Be32ToH(PUInt32(pp)^);
        Inc(pp, 4);
        Inc(p, Be32ToH(PUInt32(pp)^));
        if PChar(@p[len]) > @buf[size] then Exit;
        SetString(ws, nil, len div 2);
        for i := 1 to len div 2 do
          ws[i] := WideChar(Byte(p[i*2-1]) or (Byte(p[i*2-2]) shl 8));
        Result := ws;
        Exit;
      end;

      Exit;
    end;
  end;

var
  item, ref: PPropertyItem;
  itemsize: UINT;
  itemname, ori: string;
begin
  try
    itemsize := GetPropertyItem(Image, Id, item);
    if item = nil then
    begin
      Result := '';
      Exit;
    end;

    case id of
      PropertyTagGpsLatitude: itemname := 'GpsLatitude:'#9;
      PropertyTagGpsLongitude: itemname := 'GpsLongitude:'#9;
      PropertyTagImageDescription: itemname := 'ImageDescription:'#9;
      PropertyTagEquipMake: itemname := 'EquipMake:'#9;
      PropertyTagEquipModel: itemname := 'EquipModel:'#9;
      PropertyTagOrientation: itemname := 'Orientation:'#9;
      PropertyTagSoftwareUsed: itemname := 'SoftwareUsed:'#9;
      PropertyTagDateTime: itemname := 'DateTime:'#9#9;
      PropertyTagArtist: itemname := 'Artist:'#9#9;
      PropertyTagCopyright: itemname := 'Copyright:'#9#9;
      PropertyTagICCProfile: itemname := 'ICCProfile:'#9#9;
    else
      itemname := '0x'+IHex(id)+#9#9;
    end;

    case id of
      PropertyTagGpsLatitude:
        begin
          if itemsize < 6 then
          begin
            Result := '';
            Exit;
          end;
          GetPropertyItem(Image, PropertyTagGpsLatitudeRef, ref);
          if ref = nil then
          begin
            Result := '';
            Exit;
          end;
          Result := FormatGpsPos(ref.value, item.value);
        end;
      PropertyTagGpsLongitude:
        begin
          if itemsize < 6 then
          begin
            Result := '';
            Exit;
          end;
          GetPropertyItem(Image, PropertyTagGpsLongitudeRef, ref);
          if ref = nil then
          begin
            Result := '';
            Exit;
          end;
          Result := FormatGpsPos(ref.value, item.value);
        end;
      PropertyTagOrientation:
        begin
          ori := IStr(PSmallInt(item.value)^) + ' (';
          case PSmallInt(item.value)^ of
            1: ori := ori + 'row0=top, col0=left';
            2: ori := ori + 'row0=top, col0=right';
            3: ori := ori + 'row0=bottom, col0=right';
            4: ori := ori + 'row0=bottom, col0=left';
            5: ori := ori + 'row0=left, col0=top';
            6: ori := ori + 'row0=right, col0=top';
            7: ori := ori + 'row0=right, col0=bottom';
            8: ori := ori + 'row0=left, col0=bottom';
          else
            ori := ori + 'unknown';
          end;
          ori := ori + ')';
          Result := ori;
        end;
      PropertyTagICCProfile:
        Result := GetICCProfileDesc(item.value, itemsize);
    else
      Result := PChar(item.value);
    end;
    Result := itemname + Result + #13#10;
  finally
    FreeMem(item);
  end;
end;

procedure TGdiplusFilter.Decode(InBuf: PChar; InBufSize: Cardinal;
    var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
const
  PropertyTags: array[0..10] of PROPID = (
    PropertyTagGpsLatitude,
    PropertyTagGpsLongitude,
    PropertyTagImageDescription,
    PropertyTagEquipMake,
    PropertyTagEquipModel,
    PropertyTagOrientation,
    PropertyTagSoftwareUsed,
    PropertyTagDateTime,
    PropertyTagArtist,
    PropertyTagCopyright,
    PropertyTagICCProfile
    );
var
  st: IStream;
  gin: PChar;
  bitmap: Pointer;
  i, w, h: Cardinal;
  rc: TGpRect;
  bd: TBitmapData;
  pf, cd, ct: Integer;
  fmt: TGUID;
begin
  gin := Pointer(GlobalAlloc(GMEM_FIXED, InBufSize));
  if gin = nil then
    raise EImgFilterError.Create('input buffer allocation failed');
  Move(InBuf^, gin^, InBufSize);
  if FAILED(CreateStreamOnHGlobal(HGLOBAL(gin), True, st)) then
    raise EImgFilterError.Create('CreateStreamOnHGlobal failed');

  GdipCheck(GdipCreateBitmapFromStreamICM(st, bitmap));
  GdipCheck(GdipGetImageWidth(bitmap, w));
  GdipCheck(GdipGetImageHeight(bitmap, h));
  GdipCheck(GdipGetImagePixelFormat(bitmap, pf));
  GdipCheck(GdipGetImageRawFormat(bitmap, fmt));

  ImgInfo.Comment := '';
  for i := 0 to High(PropertyTags) do
    ImgInfo.Comment := ImgInfo.Comment + GetProperty(bitmap, PropertyTags[i]);

  rc.X := 0;
  rc.Y := 0;
  rc.Width := w;
  rc.Height := h;
  GdipCheck(GdipBitmapLockBits(bitmap, rc, ImageLockModeRead,
    PixelFormat24bppRGB, bd));

  OutBufSize := Cardinal(Abs(bd.Stride))*h;
  OutBuf := Alloc(OutBufSize);
  if OutBuf = nil then
    raise EImgFilterError.Create('output buffer allocation failed');

  if bd.Stride < 0 then
    Move(bd.Scan0^, OutBuf^, OutBufSize)
  else
    for i := 0 to h-1 do
      Move(PChar(bd.Scan0)[i*Cardinal(bd.Stride)],
          OutBuf[OutBufSize-(i+1)*Cardinal(bd.Stride)], bd.Stride);
  GdipCheck(GdipBitmapUnlockBits(bitmap, bd));
  GdipCheck(GdipDisposeImage(bitmap));

  ImgInfo.Width := w;
  ImgInfo.Height := h;
  ImgInfo.Components := 3;
  if IsEqualGUID(fmt, ImageFormatGIF) then
    ImgInfo.Bpp := Ord(InBuf[10]) and 7 + 1
  else if IsEqualGUID(fmt, ImageFormatPNG) then
  begin
    cd := Ord(InBuf[24]);
    ct := Ord(InBuf[25]);
    case ct of
      0: ImgInfo.Bpp := cd; // grayscale
      2: ImgInfo.Bpp := cd * 3; // R,G,B
      3: ImgInfo.Bpp := cd; // palette index
      4: ImgInfo.Bpp := cd; // grayscale and alpha
      6: ImgInfo.Bpp := cd * 3; // R,G,B and alpha
    else
      ImgInfo.Bpp := 0; // invalid color type
    end;
  end else
    ImgInfo.Bpp := pf shr 8 and $ff;
  ImgInfo.FormatName := '';
end;

procedure GdiplusSaveImage(const FileName: WideString;
    Format: TGdiplusSaveImageFormat; const BitmapInfo: Pointer;
    BitmapData: PChar);
var
  bitmap: Pointer;
  clsid: TGUID;
begin
  GdipCheck(GdipCreateBitmapFromGdiDib(BitmapInfo, BitmapData, bitmap));
  case Format of
    gsifPNG: clsid := CLSID_EncoderPNG;
    gsifJPEG: clsid := CLSID_EncoderJPEG;
  end;
  GdipCheck(GdipSaveImageToFile(bitmap, PWideChar(FileName), clsid, nil));
  GdipCheck(GdipDisposeImage(bitmap));
end;

var
  gsi: TGdiplusStartupInput;
  token: ULONG;

initialization
  CoInitialize(nil);
  gsi.GdiplusVersion := 1;
  GdiplusEnabled :=
      Assigned(GdiplusStartup) and (GdiplusStartup(token, gsi, nil) = gpstOk);

finalization
  if GdiplusEnabled then
    GdiplusShutdown(token);
  CoUninitialize;

end.
