unit SpiFilter;

interface

uses Windows, spi_api, ImgFilter;

type
  TArchiveInfo = record
    FileName: string;
    FileSize: Cardinal;
    CompSize: Cardinal;
    DateTime: Int64;
    RawData: Pointer;
  end;

  TArchive = class(TObject)
  private
    FPluginName: string;
    FPluginFileName: WideString;
    FFileName: string;
    Handle: THandle;
    HInfo: HLOCAL;
    SpiFileInfoTop: PFileInfo;
    SpiFileInfo: PFileInfo;
    GetPluginInfo: TGetPluginInfo;
    GetArchiveInfo: TGetArchiveInfo;
    GetFile: TGetFile;
    function GetInfo(const Name: string): TArchiveInfo;
  public
    constructor Create(const PluginFileName: WideString);
    destructor Destroy; override;
    procedure Open(const AFileName: WideString);
    function EachName(var Name: string): Boolean;
    procedure Extract(const Name: string; Buf: PChar; BufSize: Cardinal);
    property FileName: string read FFileName;
    property Info[const Name: string]: TArchiveInfo read GetInfo; default;
    property PluginName: string read FPluginName;
    property PluginFileName: WideString read FPluginFileName;
  end;

  TSpiFilter = class(TImgFilter)
  private
    FExtInfo: string;
    FFileName: WideString;
    GetPluginInfo: TGetPluginInfo;
    IsSupported: TIsSupported;
    GetPictureInfo: TGetPictureInfo;
    GetPicture: TGetPicture;
    ConfigurationDlg: TConfigurationDlg;
  public
    Handle: THandle;
    destructor Destroy; override;
    procedure LoadPlugin(const PluginFileName: WideString; UseExtInfo: Boolean);
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    function TestSpi(const AFileName: string; Buf: PChar; BufSize: Cardinal):
        Boolean;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal;
        var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
        override;
    function IsArchive: Boolean;
    function HasConfigDlg: Boolean;
    procedure ShowConfigDlg(hWnd: HWND);
    property FileName: WideString read FFileName;
    property ExtInfo: string read FExtInfo;
  end;


implementation

uses NLib, NSys, ifdib;

const
  SpiErrMsg: array[0..11] of string = (
    'not implemented',
    '',
    'user cancel',
    'unknown format',
    'broken data',
    'allocate memory error',
    'memory error',
    'file read error',
    'cannot create window',
    'internal error',
    'file write error',
    'end of file error'
  );

function ShortFileName(const S: WideString): string;
var
  len: DWORD;
  buf: WideString;
begin
  len := GetShortPathNameW(PWideChar(S), nil, 0);
  SetLength(buf, len-1);
  GetShortPathNameW(PWideChar(S), PWideChar(buf), len);
  Result := buf;
end;

procedure TSpiFilter.LoadPlugin(const PluginFileName: WideString;
    UseExtInfo: Boolean);
const
  maxexts = 100;
var
  buf: array[0..255] of Char;
  i: Integer;
begin
  Handle := LoadLibraryW(PWideChar(PluginFileName));
  if Handle <> 0 then
  begin
    FFileName := PluginFileName;
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    IsSupported := GetProcAddress(Handle, 'IsSupported');
    GetPictureInfo := GetProcAddress(Handle, 'GetPictureInfo');
    GetPicture := GetProcAddress(Handle, 'GetPicture');
    ConfigurationDlg := GetProcAddress(Handle, 'ConfigurationDlg');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FName := buf;

      FExtInfo := '';
      if UseExtInfo then
        for i := 0 to maxexts-1 do
        begin
          if GetPluginInfo(2*i+2, buf, SizeOf(buf)) = 0 then
            Break;
          FExtInfo := FExtInfo + buf + ';';
        end;
    end;
  end;
end;

destructor TSpiFilter.Destroy;
begin
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TSpiFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
begin
  Result := TestSpi('', Buf, BufSize);
end;

function TSpiFilter.TestSpi(const AFileName: string; Buf: PChar;
    BufSize: Cardinal): Boolean;
var
  tmpbuf: array[0..2047] of Char;
  root, ext: WideString;
  info: string;
begin
  if not Assigned(IsSupported) then
  begin
    Result := False;
    Exit;
  end;

  if ExtInfo <> '' then
  begin
    if AFileName = '' then // stdin
    begin
      Result := False;
      Exit;
    end;
    SplitExt(AFileName, root, ext);
    ext := StrUpper(ext);
    info := StrUpper(ExtInfo);
    if FindBytes(info[1], Length(info), '*'+ext+';') = -1 then
    begin
      Result := False;
      Exit;
    end;
  end;

  // Call IsSupported() with copyed buffer. IsSupprted() may destroy buffer.
  if BufSize > SizeOf(tmpbuf) then
    BufSize := SizeOf(tmpbuf);
  FillChar(tmpbuf, SizeOf(tmpbuf), 0);
  Move(Buf^, tmpbuf, BufSize);
  Result := IsSupported(PChar(AFileName), DWORD(@tmpbuf)) <> 0;
end;

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

procedure SpiCheck(Code: Integer; const Name: string);
var
  msg: string;
  msgidx: Integer;
begin
  if Code = 0 then
    Exit;
  msgidx := Code + 1;
  if (msgidx >= 0) and (msgidx <= High(SpiErrMsg)) then
    msg := SpiErrMsg[msgidx] + ' ('+IStr(Code)+')'
  else
    msg := 'error ' + IStr(Code);
  raise EImgFilterError.Create(Name+': '+msg);
end;

procedure TSpiFilter.Decode(InBuf: PChar; InBufSize: Cardinal;
    var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
var
  hbinfo, hbm: HLOCAL;
  size: Integer;
  bmi: TMaxBitmapInfo;
  picinfo: TPictureInfo;
begin
  // For plugin which do not set all picinfo paramateres.
  FillChar(picinfo, SizeOf(picinfo), 0);
  try
    if Assigned(GetPictureInfo) then
      SpiCheck(GetPictureInfo(InBuf, InBufSize, 1, picinfo), 'GetPictureInfo');

    with ImgInfo do
    begin
      Bpp := picinfo.colorDepth;
      FormatName := '';
      if picinfo.hInfo = 0 then
        Comment := ''
      else begin
        SetString(Comment, PChar(LocalLock(picinfo.hInfo)),
          LocalSize(picinfo.hInfo)-1); // -1: for last null character
        LocalUnLock(picinfo.hInfo);
      end;
      Components := 3; // DIB decoded by SPI is always 24bpp(3bytes).
    end;
  finally
    if picinfo.hInfo <> 0 then
      LocalFree(picinfo.hInfo);
  end;

  if not Assigned(GetPicture) then
    raise EImgFilterError.Create('GetPicture not found');

  // Call GetPicture() with dummy ProgressCallback function.
  // Some plugin may cause error if ProgressCallback is nil.
  hbinfo := 0;
  hbm := 0;
  try
    SpiCheck(GetPicture(InBuf, InBufSize, 1, hbinfo, hbm, ProgressCallback, 0),
        'GetPicture');

    size := LocalSize(hbinfo);
    if size > SizeOf(TMaxBitmapInfo) then
      size := SizeOf(TMaxBitmapInfo);
    Move(PMaxBitmapInfo(LocalLock(hbinfo))^, bmi, size);
    LocalUnLock(hbinfo);

    ImgInfo.Width := bmi.bmiHeader.biWidth;
    ImgInfo.Height := Abs(bmi.bmiHeader.biHeight);

    OutBufSize := ((ImgInfo.Width * 3 + 3) and not 3) * ImgInfo.Height;
    OutBuf := Alloc(OutBufSize);
    if OutBuf = nil then
      raise EImgFilterError.Create('output buffer allocation failed');

    DecodeDIB(LocalLock(hbm), LocalSize(hbm), bmi, OutBuf, OutBufSize);
    LocalUnLock(hbm);
  finally
    if hbm <> 0 then
      LocalFree(hbm);
  end;
end;

function TSpiFilter.IsArchive: Boolean;
var
  ver: array[0..15] of Char;
begin
  FillChar(ver, SizeOf(ver), 0);
  if Assigned(GetPluginInfo) then
    GetPluginInfo(0, ver, SizeOf(ver));
  Result := Copy(ver, 1, 4) = '00AM';
end;

function TSpiFilter.HasConfigDlg: Boolean;
begin
  Result := Assigned(ConfigurationDlg);
end;

procedure TSpiFilter.ShowConfigDlg(hWnd: HWND);
begin
  SpiCheck(ConfigurationDlg(hWnd, 1), 'ConfigurationDlg');
end;


{ TArchive }

constructor TArchive.Create(const PluginFileName: WideString);
var
  buf: array[0..255] of Char;
begin
  Handle := LoadLibraryW(PWideChar(PluginFileName));
  if Handle <> 0 then
  begin
    FPluginFilename := PluginFileName;
    GetArchiveInfo := GetProcAddress(Handle, 'GetArchiveInfo');
    GetFile := GetProcAddress(Handle, 'GetFile');
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FPluginName := buf;
    end;
  end;
end;

procedure TArchive.Open(const AFileName: WideString);
begin
  FFileName := ShortFileName(AFileName);
  if not Assigned(GetArchiveInfo) then
    raise EImgFilterError.Create('GetArchiveInfo not found');
  SpiCheck(GetArchiveInfo(PChar(FFileName), 0, 0, HInfo), 'GetArchiveInfo');
  SpiFileInfoTop := LocalLock(HInfo);
  SpiFileInfo := SpiFileInfoTop;
end;

destructor TArchive.Destroy;
begin
  if HInfo <> 0 then
  begin
    LocalUnlock(HInfo);
    LocalFree(HInfo);
  end;
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TArchive.GetInfo(const Name: string): TArchiveInfo;
var
  spiinfo: PFileInfo;
begin
  spiinfo := SpiFileInfoTop;
  while spiinfo.method[0] <> #0 do
    if Name = spiinfo.path + spiinfo.filename then
    begin
      Result.FileName := Name;
      Result.FileSize := spiinfo.filesize;
      Result.CompSize := spiinfo.compsize;
      Result.DateTime := spiinfo.timestamp;
      Result.RawData := spiinfo;
      Exit;
    end else
      Inc(spiinfo);
  Result.FileName := '';
  Result.RawData := nil;
end;

function TArchive.EachName(var Name: string): Boolean;
begin
  Result := SpiFileInfo.method[0] <> #0;
  if Result then
    Name := SpiFileInfo.path + SpiFileInfo.filename;
  Inc(SpiFileInfo);
end;

procedure TArchive.Extract(const Name: string; Buf: PChar; BufSize: Cardinal);
var
  info: PFileInfo;
  hdst: HLOCAL;
  p: PChar;
  fsize: UINT;
begin
  info := GetInfo(Name).RawData;
  if info = nil then
    raise EImgFilterError.Create('Archive file info not found');

  if not Assigned(GetFile) then
    raise EImgFilterError.Create('GetFile not found');
  SpiCheck(
      GetFile(PChar(FFileName), info.position, PChar(@hdst), $100, nil,0),
      'GetFile');
  p := LocalLock(hdst);
  fsize := LocalSize(hdst);
  if fsize > BufSize then
    fsize := BufSize;
  Move(p^, Buf^, fsize);
  LocalUnLock(hdst);
  LocalFree(hdst);
end;

end.
